import gc
import sys
import time
import uuid
from statistics import mean

from .backends.interface import Backend


async def _get_latency(func, **kwargs) -> float:
    start = time.perf_counter()
    await func(**kwargs)
    return time.perf_counter() - start


async def _get_average_latency(func, iters: int = 100, **kwargs):
    latency = []
    for _ in range(iters):
        latency.append(await _get_latency(func, **kwargs))
    return {"min": min(latency), "max": max(latency), "mean": mean(latency)}


async def _get_results(backend: Backend, key: str, iters=100):
    return await _get_average_latency(backend.get, key=key, iters=iters)


async def _set_results(backend: Backend, key: str, iters=100):
    return await _get_average_latency(backend.set, key=key, value="1", expire=10, iters=iters)


async def check_speed(backend: Backend, iters: int = 100):
    key = f"check:{str(uuid.uuid4())}"
    return {
        "set": await _set_results(backend, key, iters=iters),
        "get": await _get_results(backend, key, iters=iters),
    }


def _get_obj_size(obj) -> int:
    marked = {id(obj)}
    obj_q = [obj]
    size = 0

    while obj_q:
        size += sum(map(sys.getsizeof, obj_q))

        # Lookup all the object referred to by the object in obj_q.
        # See: https://docs.python.org/3.7/library/gc.html#gc.get_referents
        all_refr = ((id(o), o) for o in gc.get_referents(*obj_q))

        # Filter object that are already marked.
        # Using dict notation will prevent repeated objects.
        new_refr = {o_id: o for o_id, o in all_refr if o_id not in marked and not isinstance(o, type)}

        # The new obj_q will be the ones that were not marked,
        # and we will update marked with their ids so we will
        # not traverse them again.
        obj_q = new_refr.values()
        marked.update(new_refr.keys())

    return size



import asyncio
import logging
import socket
from typing import Any, Optional, Union

from aioredis import ConnectionsPool as __ConnectionsPool
from aioredis import Redis as Redis_
from aioredis import RedisError, create_pool

from ..serialize import PickleSerializerMixin
from .interface import Backend

__all__ = "SafeRedis"
logger = logging.getLogger(__name__)
_UNLOCK = """
if redis.call("get",KEYS[1]) == ARGV[1] then
    return redis.call("del",KEYS[1])
else
    return 0
end
"""
# pylint: disable=arguments-differ
# pylint: disable=abstract-method


class _Redis(Redis_):
    def __init__(self, address, safe=False, **kwargs):
        self._address = address
        kwargs.pop("local_cache", None)
        kwargs.pop("prefix", None)
        self._kwargs = kwargs
        self._pool_or_conn: Optional[_ConnectionsPool] = None
        self._sha = {}
        self._safe = safe

    @property
    def is_init(self):
        return bool(self._pool_or_conn)

    async def init(self):
        try:
            self._pool_or_conn = await create_pool(address=self._address, pool_cls=_ConnectionsPool, **self._kwargs)
        except Exception:
            if not self._safe:
                raise

    def get_many(self, *keys: str):
        return self.mget(keys[0], *keys[1:])

    def clear(self):
        return self.flushdb()

    async def set(self, key: str, value: Any, expire: Union[None, float, int] = None, exist=None):
        if exist is True:
            exist = Redis_.SET_IF_EXIST
        elif exist is False:
            exist = Redis_.SET_IF_NOT_EXIST
        pexpire = None
        if isinstance(expire, float):
            pexpire = int(expire * 1000)
            expire = None
        return await super().set(key, value, expire=expire, pexpire=pexpire, exist=exist)

    def get_expire(self, key: str) -> int:
        return self.ttl(key)

    def set_lock(self, key: str, value, expire):
        pexpire = None
        if isinstance(expire, float):
            pexpire = int(expire * 1000)
            expire = None
        return super().set(key, value, expire=expire, pexpire=pexpire, exist=self.SET_IF_NOT_EXIST)

    async def is_locked(self, key: str, wait=None, step=0.1):
        if wait is None:
            return await self.exists(key)
        while wait > 0.0:
            if not await self.exists(key):
                return False
            wait -= step
            await asyncio.sleep(step)
        return True

    async def unlock(self, key, value):
        if "UNLOCK" not in self._sha:
            self._sha["UNLOCK"] = await self.script_load(_UNLOCK.replace("\n", " "))
        return await self.evalsha(self._sha["UNLOCK"], keys=[key], args=[value])

    def delete(self, key):
        return self.unlink(key)

    def exists(self, key) -> bool:
        return super().exists(key)

    async def keys_match(self, pattern: str):
        cursor = b"0"
        while cursor:
            cursor, keys = await self.scan(cursor, match=pattern, count=100)
            for key in keys:
                yield key

    async def delete_match(self, pattern: str):
        if "*" not in pattern:
            return await self.unlink(pattern)
        keys = []
        async for key in self.keys_match(pattern):
            keys.append(key)
        if keys:
            return await self.unlink(keys[0], *keys[1:])

    async def get_size(self, key: str) -> int:
        return int(await self.execute(b"MEMORY", b"USAGE", key))

    async def get(self, key: str, **kwargs) -> Any:
        return await super().get(key=key, **kwargs)

    async def execute(self, command, *args, **kwargs):
        try:
            return await super().execute(command, *args, **kwargs)
        except (RedisError, socket.gaierror, OSError, asyncio.TimeoutError, AttributeError):
            if not self._safe or command.lower() == b"ping":
                raise
            logger.error("Redis down on command %s", command)
            if command.lower() in [b"unlink", b"del", b"memory", b"ttl"]:
                return 0
            if command.lower() == b"scan":
                return [0, []]
            return None


class Redis(PickleSerializerMixin, _Redis, Backend):
    pass


class _ConnectionsPool(__ConnectionsPool):
    async def _create_new_connection(self, address):
        conn = await super()._create_new_connection(address)
        await conn.execute(b"CLIENT", b"SETNAME", b"cachews")
        return conn



from typing import Tuple

from cashews.key import get_template_for_key

from .redis import Redis


class IndexRedis(Redis):
    """
    Backend based on redis backend to have ability to invalidate the group of keys based on index

    for example index == user_uid
    hkey user:{user_uid} {"info": ..., "wallets": ...., "whatever": .....}

    accounts:{user_uid} -> index_name:{user_uid} {"accounts": ....}
    """

    def __init__(self, index_name, index_field, *args, **kwargs):
        self._index_name = index_name
        self._index_field = index_field
        super().__init__(*args, **kwargs)

    def _get_index_and_key(self, key) -> Tuple[str, str]:
        template, groups = get_template_for_key(key)
        if not template or self._index_field not in groups:
            return "", key
        key = key.replace(":" + groups[self._index_field], "").replace(groups[self._index_field], "")
        return self._index_name + ":" + groups[self._index_field], key

    def set(self, key: str, value, *args, **kwargs):
        index, key = self._get_index_and_key(key)
        if index:
            return self.hset(index, key, value)
        return super().set(key, value, *args, **kwargs)

    def get(self, key: str, *args, **kwargs):
        index, key = self._get_index_and_key(key)
        if index:
            return self.hget(index, key)
        return super().get(key)

    def delete(self, key):
        index, key = self._get_index_and_key(key)
        if index:
            return self.hdel(index, key)
        return super().delete(key)

    def delete_match(self, pattern: str):
        index, key = self._get_index_and_key(pattern)
        if index:
            return self.unlink(index)
        return super().delete_match(pattern)



import os

import pytest

from cashews.backends.index import IndexRedis
from cashews.key import register_template

pytestmark = pytest.mark.asyncio
REDIS_TESTS = bool(os.environ.get("USE_REDIS"))


@pytest.fixture(name="cache")
async def _cache():
    redis = IndexRedis(address="redis://", hash_key=None, index_field="user", index_name="test")
    await redis.init()
    await redis.clear()
    return redis


@pytest.mark.skipif(not REDIS_TESTS, reason="only for redis")
async def test_set(cache):
    register_template(test_set, "key:{user}:{account}")
    await cache.set("key:jon:10", b"val")
    assert await cache.hget("test:jon", "key:10") == b"val"


@pytest.mark.skipif(not REDIS_TESTS, reason="only for redis")
async def test_get(cache):
    register_template(test_get, "key:{user}:{account}")
    await cache.hset("test:jon", "key:10", b"val")

    assert await cache.get("key:jon:10") == b"val"


@pytest.mark.skipif(not REDIS_TESTS, reason="only for redis")
async def test_get_set_no_template(cache):
    await cache.set("1:jon:10", b"val")
    assert await cache.get("1:jon:10") == b"val"


@pytest.mark.skipif(not REDIS_TESTS, reason="only for redis")
async def test_delete_match(cache):
    register_template(test_get, "key:{user}:{account}")
    await cache.hset("test:jon", "key:10", b"val")

    await cache.delete_match("key:jon:*")
    assert await cache.get("key:jon:10") is None


@pytest.mark.skipif(not REDIS_TESTS, reason="only for redis")
async def test_delete(cache):
    register_template(test_get, "key:{user}:{account}")
    await cache.hset("test:jon", "key:10", b"val")

    await cache.delete("key:jon:10")
    assert await cache.get("key:jon:10") is None
